﻿using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using HeuristicLab.Algorithms.PushGP.Expressions;
using HeuristicLab.Algorithms.PushGP.Stack;

namespace HeuristicLab.Algorithms.PushGP.Interpreter
{
    public class PushGPInterpreter : IInterpreterService
    {
        private static Parser parser = new Parser();
        private Task currentTask;

        public PushGPInterpreter(Configuration config = null)
        {
            this.Configuration = config ?? new Configuration();

            this.CodeStack = new PushGPStack<Expression>();
            this.ExecStack = new PushGPStack<Expression>();
            this.NameStack = new PushGPStack<string>();
            this.BooleanStack = new PushGPStack<bool>();
            this.IntegerStack = new PushGPStack<long>();
            this.FloatStack = new PushGPStack<double>();
        }

        public Configuration Configuration { get; }

        public bool IsPaused { get; private set; } = false;
        public bool IsAborted { get; private set; } = false;
        public bool IsRunning => this.ExecStack.Count > 0 && !IsPaused && !IsAborted;
        public bool IsCompleted => this.ExecStack.Count == 0;
        public bool CanStep => !IsCompleted && !IsAborted && IsPaused;

        public IStack<Expression> CodeStack { get; }
        public IStack<Expression> ExecStack { get; }
        public IStack<string> NameStack { get; }
        public IStack<bool> BooleanStack { get; }
        public IStack<long> IntegerStack { get; }
        public IStack<double> FloatStack { get; }

        public IDictionary<string, Expression> CustomExpressions { get; } = new Dictionary<string, Expression>();

        public void Reset()
        {
            this.IsAborted = false;
            this.IsPaused = false;
            this.currentTask = null;

            this.ExecStack.Clear();
            this.CodeStack.Clear();
            this.NameStack.Clear();
            this.BooleanStack.Clear();
            this.IntegerStack.Clear();
            this.FloatStack.Clear();
        }

        public async void AbortAndReset()
        {
            await this.AbortAsync();
            this.Reset();
        }

        public void Interprete(string code)
        {
            var program = Encode(code);
            this.Interprete(program);
        }

        public Task InterpreteAsync(string code, CancellationToken token = default(CancellationToken))
        {
            var program = Encode(code);

            this.currentTask = this.InterpreteAsync(program, token);

            return currentTask;
        }

        public Task InterpreteAsync(Expression program, CancellationToken token = default(CancellationToken))
        {
            this.currentTask = Task.Run(() => this.Interprete(program), token);

            return currentTask;
        }

        public void Interprete(Expression program)
        {
            this.ExecStack.Push(program);
            this.CodeStack.Push(program);

            // run top expression
            this.ExecStack.Pop().Eval(this);

            this.Interprete();
        }

        private void Interprete()
        {
            while (IsRunning)
            {
                this.DoStep();
            }

            if (this.Configuration.TopLevelPushCode)
            {
                this.CodeStack.Pop();
            }
        }

        private void DoStep()
        {
            var expression = this.ExecStack.Pop();
            var isNotCodeOp = !expression.OpCode.IsCodeOp();

            if (isNotCodeOp)
            {
                this.CodeStack.Push(expression);
            }

            expression.Eval(this);

            if (isNotCodeOp)
            {
                CodeStack.Pop();
            }
        }

        private Task InterpreteAsync()
        {
            this.currentTask = Task.Run(() =>
            {
                this.Interprete();
            });

            return this.currentTask;
        }

        public async Task AbortAsync()
        {
            if (this.IsAborted || this.IsCompleted)
                return;

            this.IsAborted = true;

            if (this.currentTask != null)
            {
                await this.currentTask;
            }
        }

        public async Task PauseAsync()
        {
            if (this.IsPaused || this.IsCompleted)
                return;

            this.IsPaused = true;

            if (this.currentTask != null)
            {
                await this.currentTask;
            }
        }

        public Task ResumeAsync()
        {
            if (this.IsPaused)
            {
                this.IsPaused = false;
                return this.InterpreteAsync();
            }

            return this.currentTask;
        }

        public void Step()
        {
            if (CanStep)
            {
                DoStep();
            }
        }

        public static Expression Encode(string code)
        {
            return parser.Parse(code);
        }

        public static Task<Expression> EncodeAsync(string code, CancellationToken token = default(CancellationToken))
        {
            return Task.Run(() => Encode(code), token);
        }

        public static string Decode(Expression expression)
        {
            return string.Concat(expression.ToString());
        }

        public static Task<string> DecodeAsync(Expression expression, CancellationToken token = default(CancellationToken))
        {
            return Task.Run(() => Decode(expression), token);
        }

        public void PrintStacks()
        {
            this.PrintStack("EXEC", this.ExecStack);
            this.PrintStack("CODE", this.CodeStack);
            this.PrintStack("NAME", this.NameStack);
            this.PrintStack("BOOLEAN", this.BooleanStack);
            this.PrintStack("FLOAT", this.FloatStack);
            this.PrintStack("INTEGER", this.IntegerStack);
        }

        private void PrintStack<T>(string name, IStack<T> stack)
        {
            System.Console.WriteLine($"--- {name} ---\n{stack}\n---------------------------------------------------");
        }
    }
}
